import threading
import datetime
import schedule
import asyncio
import logging
import time
from flask import Flask, json, request, Response
from werkzeug.wsgi import DispatcherMiddleware
from .utils import AttrDict, TaggedDict, mangle_name, join_url, _APIWrapper
from .dispatch import Dispatcher

LOG = logging.getLogger("idiotic.init")

# Monkeypatch schedule so nobody has to deal with it directly
def __sched_job_do_once(self, func, *args, **kwargs):
    self.do(lambda: func(*args, **kwargs) and schedule.CancelJob or schedule.CancelJob)

def __sched_job_do_now(self, func, *args, **kwargs):
    func(*args, **kwargs)
    self.do(func, *args, **kwargs)

schedule.Job.do_once = __sched_job_do_once
schedule.Job.do_now = __sched_job_do_now

instance = None

distrib_types = {}
persist_types = {}

def _register_distrib_type(name, distrib):
    global distrib_types
    distrib_types[name] =  distrib

def _register_persistence(name, cls):
    global persist_types
    persist_types[name] = cls

def _register_scene(name, scene):
    instance._register_scene(name, scene)

class Idiotic:
    def __init__(self, config=None, name="idiotic"):
        if config is None:
            config = {}
        self.config = config
        self.name = name

        self.items = TaggedDict()
        self.scenes = TaggedDict()
        self.modules = AttrDict()
        self.scheduler = schedule.Scheduler()
        self.dispatcher = Dispatcher()
        self.persist_instance = None
        self.distribution = None
        self.distrib_thread = None
        self._root_api = Flask(__name__)
        self._apis = {}
        self.api = None

    @asyncio.coroutine
    def run_scheduled_jobs(self):
        while True:
            try:
                runnable_jobs = sorted((job for job in self.scheduler.jobs if job.should_run))
                if len(runnable_jobs):
                    for job in runnable_jobs:
                        # i'm sorry this is kind of terrible
                        # but dammit, it works
                        ret = yield from asyncio.coroutine(job.job_func)()

                        while asyncio.iscoroutine(ret):
                            ret = yield from ret

                        job.last_run = datetime.datetime.now()
                        job._schedule_next_run()

                        if isinstance(ret, schedule.CancelJob):
                            self.scheduler.cancel_job(job)
                else:
                    yield from asyncio.sleep(1)
            except:
                LOG.exception("Exception in scheduler.run_pending()")
                yield from asyncio.sleep(1)

    def _register_item(self, item):
        self.items._set(item.name, item)
        item.idiotic = self

    def _register_scene(self, name, scene):
        self.scenes._set(name, scene)
        scene.idiotic = self

    def _register_module(self, module, assets=None):
        name = mangle_name(getattr(module, "MODULE_NAME", module.__name__))

        mod_conf = self.config.get("modules", {}).get(name, {})

        if mod_conf.get("disable", False):
            LOG.info("Module {} is disabled; skipping registration".format(name))
            return

        base = mod_conf.get("api_base", "/api/module/" + name)

        if base == '/':
            mod_api = self._root_api
        elif base in self._apis:
            mod_api = self._apis[base]
        else:
            mod_api = Flask(name)

        self._apis[mod_conf.get("api_base", "/api/module/" + name)] = mod_api

        if hasattr(module, "configure"):
            LOG.info("Configuring module {}".format(name))
            module.configure(
                mod_conf,
                _APIWrapper(mod_api, module, '/'),
                assets
            )

        if hasattr(module, "start"):
            LOG.info("Starting module {}".format(name))
            asyncio.get_event_loop().call_soon(module.start)

        self.modules._set(name, module)

    def _finalize_api(self):
        self.api = DispatcherMiddleware(self._root_api, self._apis)

        return self.api

    def _register_builtin_module(self, module, assets=None):
        name = mangle_name(getattr(module, "MODULE_NAME", module.__name__))

        if hasattr(module, "configure"):
            LOG.info("Configuring system module {}".format(name))
            module.configure(
                self.config,
                self.config.get(name, {}),
                _APIWrapper(self._root_api, module, '/'),
                assets
            )

        self.modules._set(name, module)

    def augment_module(self, module):
        setattr(module, "context", self)
        setattr(module, "modules", self.modules)
        setattr(module, "items", self.items)
        setattr(module, "scenes", self.scenes)
        setattr(module, "scheduler", self.scheduler)

    def _send_event(self, evt):
        LOG.debug("_send_event!")
        self.distribution.send(event.pack_event(evt))

    def _recv_event(self, evt):
        LOG.debug("_recv_event!")
        self.dispatcher.dispatch(event.unpack_event(evt, self.modules))

    def _start_distrib(self, dist, host, conf):
        try:
            dist_cls = distrib_types[dist]
        except NameError as e:
            raise NameError("Could not find distribution method {}".format(dist), e)
        self.distribution = dist_cls(host, conf)
        self.distribution.connect()

        self.dispatcher.bind(_send_event, utils.Filter(not_hasattr='_remote'))
        self.distribution.receive(_recv_event)

        thread = threading.Thread(target=self.distribution.run, daemon=True)
        thread.start()

    def _stop_distrib(self):
        if self.distribution:
            self.distribution.disconnect()

        if self.distrib_thread:
            self.distrib_thread.join()

    def _record_state_change(self, evt):
        if evt and evt.item:
            self.persist_instance.append_item_history(evt.item, evt.time, evt.new, kind="state")

    def _start_persistence(self, persist, conf):
        persist_cls = persist_types[persist]
        self.persist_instance = persist_cls(conf)
        self.persist_instance.connect()
        self.dispatcher.bind(self._record_state_change, utils.Filter(type=event.StateChangeEvent, kind="after"))

        for item in self.items.all():
            history = list(self.persist_instance.get_item_history(item))
            if len(history):
                item._state = history[-1][0]

            if hasattr(item, "state_history"):
                for state in history:
                    item.state_history.record(*state)

    def _stop_persistence(self):
        if self.persist_instance:
            self.persist_instance.sync()
            self.persist_instance.disconnect()



import json

def mangle_name(name):
    return ''.join((x for x in name.lower().replace(" ", "_") if x.isalnum() or x=='_')) if name else ""

class IdioticEncoder(json.JSONEncoder):
    def __init__(self, *args, depth=-1, **kwargs):
        super().__init__(*args, **kwargs)
        self.depth = depth

    def default(self, obj):
        if obj is None:
            return "null"
        if hasattr(obj, "pack"):
            return obj.pack()
        else:
            return json.JONEncoder.default(self, obj)



from . import _register_persistence

class PersistenceType(type):
    def __init__(cls, name, bases, attrs):
        super(PersistenceType, cls).__init__(name, bases, attrs)
        if name != "Persistence":
            _register_persistence(getattr(cls, "NAME", name.lower()), cls)

class NotConnected(Exception):
    pass

class Persistence(metaclass=PersistenceType):
    def __init__(self, config):
        pass

    def __enter__(self):
        self.connect()

    def __exit__(self, type, value, traceback):
        self.disconnect()

    def connect(self):
        """Establish a connection to the database, if necessary."""
        pass

    def disconnect(self):
        """Close the connection to the database, if necessary."""
        pass

    def create(self):
        """Create and initialize the database, if it does not already exist.
May be called automatically by the persistence engine if needed.

        """
        pass

    def version(self):
        """The version of the database when it was created or last updated.
"""
        pass

    def upgrade(self, old_version, new_version):
        """Convert the database from an old version to a new version."""
        pass

    def sync(self):
        """Write any uncommitted data to the underlying database engine."""
        pass

    def purge(self):
        """Delete any data that does not meet retention requirements."""
        pass

    def get_item_history(self, item, kind="state", since=None, count=-1):
        """Retrieve the history for the given item."""
        pass

    def append_item_history(self, item, time, value, kind="state", extra=None):
        pass

    def set_item_history(self, item, histories, kind="state"):
        pass



from . import base
import logging
import socket
import struct
import queue

PACKET_HEAD = b"ID10T"
ID_EVENT = 1
ID_DISCOVERY = 2

EVENT = 1
DISCOVERY = 2
RESPONSE = 3

FORMAT = {
    EVENT: "{}s",
    DISCOVERY: "H{}s",
    RESPONSE: "H{}s",
}

HEADER_FORMAT = "!5sBI"
HEADER_LEN = struct.calcsize(HEADER_FORMAT)

LOG = logging.getLogger("idiotic.distrib.udp")

class UDPItem(base.RemoteItem):
    pass

class UDPModule(base.RemoteModule):
    pass

class UDPNeighbor(base.Neighbor):
    def __init__(self, name, host, port):
        self.name = name
        self.host = host
        self.port = port
        self.modules = []
        self.items = []

class UDPTransportMethod(base.TransportMethod):
    NEIGHBOR_CLASS = UDPNeighbor
    MODULE_CLASS = UDPModule
    ITEM_CLASS = UDPItem
    NAME = "udp"

    def __init__(self, hostname, config):
        self.hostname = hostname

        config = config or {}

        self.listen_port = config.get("port", 28300)
        self.listener = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.listener.settimeout(5)
        self.listener.bind(('', self.listen_port))

        self.sender = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sender.bind(('', 0))
        self.sender.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        self.incoming = queue.Queue()
        self.neighbor_dict = {}

        for connection in config.get("connect", []):
            if 'name' in connection and 'host' in connection:
                self.neighbor_dict[connection['name']] = UDPNeighbor(
                    connection['name'],
                    connection['host'],
                    connection.get('port',
                                   self.listen_port))

        self.running = False

    def _encode_packet(self, kind, *data):
        strlens = [len(s) for s in data if isinstance(s, str) or isinstance(s, bytes)]
        msg_len = struct.calcsize('!' + FORMAT[kind].format(*strlens))

        return struct.pack(HEADER_FORMAT + FORMAT[kind].format(*strlens),
                           PACKET_HEAD, kind, msg_len,
                           *[s.encode('UTF-8') if type(s) is str else s for s in data])

    def _decode_packet(self, data):
        head, kind, data_len = struct.unpack_from(HEADER_FORMAT, data)

        if head != PACKET_HEAD:
            raise ValueError("Invalid packet header '{}'".format(head))

        if kind == EVENT:
            tup = struct.unpack_from('!' + FORMAT[EVENT].format(data_len), data, HEADER_LEN)
        elif kind == DISCOVERY or kind == RESPONSE:
            port, host = struct.unpack_from('!' + FORMAT[DISCOVERY].format(
                data_len - struct.calcsize(FORMAT[DISCOVERY][:1])), data, HEADER_LEN)
            tup = port, host.decode('UTF-8')
        return kind, tup

    def _send_discovery(self, target='<broadcast>', port=None, response=False):
        if port is None:
            port = self.listen_port

        LOG.info("Sending discovery message to ({}, {})".format(target, port))
        try:
            self.sender.sendto(self._encode_packet(RESPONSE if response else DISCOVERY,
                                                   self.listen_port, self.hostname),
                               (target, port))
        except OSError:
            LOG.warn("Unable to send UDP packet")

    def connect(self):
        for neighbor in self.neighbor_dict.values():
            self._send_discovery(neighbor.host, neighbor.port)
        self._send_discovery()

    def run(self):
        LOG.info("Starting UDP Distribution client.")
        self.running = True
        while self.running:
            try:
                data, addr = self.listener.recvfrom(2048)
                LOG.debug("Received '{}' from {}".format(data, addr))
                try:
                    kind, tup = self._decode_packet(data)
                except ValueError:
                    LOG.error("Received invalid packet from {}: {}".format(addr, data))
                if kind == DISCOVERY or kind == RESPONSE:
                    LOG.debug("Received discovery packet")

                    port, host = tup

                    if host == self.hostname and port == self.listen_port:
                        # Skip our own packets because... well... we're already
                        # connected to us...
                        continue
                    if host in self.neighbor_dict:
                        LOG.debug("Updating existing neighbor {}".format(host))
                        self.neighbor_dict[host].name = host
                        self.neighbor_dict[host].host = addr[0]
                        self.neighbor_dict[host].port = port
                    else:
                        LOG.info("Found new neighbor {} at {}".format(host, addr))
                        self.neighbor_dict[host] = UDPNeighbor(host, addr[0], port)

                    if kind != RESPONSE:
                        self._send_discovery(addr[0], port, response=True)

                elif kind == ID_EVENT:
                    LOG.debug("Received event packet")
                    event, = tup
                    self.__do_callback(event)

                else:
                    LOG.debug("Bad message kind: {}".format(kind))
            except socket.timeout:
                continue

    def stop(self):
        self.running = False

    def disconnect(self):
        pass

    def __do_callback(self, event):
        for cb in list(self.callbacks):
            cb(event)

    def send(self, event, targets=True):
        LOG.debug("Sending event {} to: {}".format(event, targets))
        if targets is True:
            targets = [('<broadcast>', self.listen_port)]
        else:
            targets = [(self.neighbor_dict[n].host, self.neighbor_dict[n].port) for n in targets
                       if n in self.neighbor_dict]

        for target in targets:
            self.sender.sendto(self._encode_packet(EVENT, event),
                               target)

    def neighbors(self):
        return list(self.neighbor_dict.values())

METHOD = UDPTransportMethod



import collections
import datetime
import bisect

Entry = collections.namedtuple('Entry', ['time', 'state'])

class History:
    def __init__(self, initial=[], maxlen=None, maxage=None):
        self.values = collections.deque(sorted((Entry(*i) for i in initial)), maxlen=maxlen)

        if isinstance(maxage, int):
            self.maxage = datetime.timedelta(seconds=maxage)
        elif isinstance(maxage, datetime.timedelta):
            self.maxage = maxage
        elif maxage is None:
            self.maxage = None
        else:
            raise ValueError("maxage must be int or timedelta")

    def cull(self):
        if self.maxage:
            pos = bisect.bisect_left(list(zip(*self.values))[0], datetime.datetime.now() - self.maxage)
            for _ in range(pos):
                self.values.popleft()

    def record(self, value, time=None):
        if time is None:
            self.values.append(Entry(datetime.datetime.now(), value))
        elif isinstance(time, datetime.datetime):
            if len(self.values) and time < self.values[-1][0]:
                raise NotImplementedError("We can't alter history!... yet....")
            else:
                self.values.append(Entry(time, value))
        else:
            raise ValueError("time must be datetime")

    def at(self, time):
        return self.values[bisect.bisect(list(zip(*self.values))[0], time)-1]

    def since(self, time):
        return list(self.values)[bisect.bisect(list(zip(*self.values))[0], time)-1:]

    def all(self):
        return list(self.values)

    def last(self, nth=None):
        if nth:
            return list(self.values)[-nth:]
        else:
            if self.values:
                return self.values[-1]
            else:
                return []

    def __len__(self):
        return len(self.values)

    def __getitem__(self, pos):
        return list(self.values)[pos]

    def __str__(self):
        return str(self.values)



from setuptools import setup, find_packages
import os

setup(
    name='idiotic',
    packages=find_packages(exclude=['etc', 'contrib']),
    version='0.1.2',
    description='Distributed home automation controller',
    long_description="""The idiotic distributed internet of things inhabitance
    controller (idiotic), aims to be an extremely extensible, capable, and most
    importantly developer-and-user-friendly solution for mashing together a wide
    assortment of existing home automation technologies into something which is
    useful as a whole.""",    
    license="""The MIT License (MIT)

    Copyright (c) 2015-2016 Hackafe and contributors

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
""",
    author='Dylan Whichard',
    author_email='dylan@whichard.com',
    url='https://github.com/umbc-hackafe/idiotic',
    keywords=[
        'home automation', 'iot', 'internet of things'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: No Input/Output (Daemon)',
        'Framework :: Flask',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Home Automation',
    ],
    install_requires=[
        'docopt>=0.6.2',
        'schedule>=0.3.2',
        'asyncio',
        'aiohttp>=0.21.2',
        'werkzeug0.11.4',
        'Flask>=0.10.1',
    ],
    data_files=[
        ('/usr/lib/systemd/system/idiotic.service', ['contrib/idiotic.service']),
        ('/etc/idiotic/', ['contrib/conf.json']),
    ],
    entry_points={
        'console_scripts': [
            'idiotic=idiotic.__main__:main',
        ]
    },
)



import asyncio

class Timer:
    def __init__(self, delay, func, *args, **kwargs):
        self.started = False
        self.completed = False
        self.canceled = False

        self.delay = delay

        self.func = func
        self.args = args
        self.kwargs = kwargs

        self.__handle = None

    def start(self):
        self.reschedule()

    def reschedule(self, delay=None):
        if delay is None:
            delay = self.delay
        else:
            self.delay = delay

        self.cancel()

        self.completed = False
        self.canceled = False
        self.__handle = asyncio.get_event_loop().call_later(delay, self.__run)
        self.started = True

    def __run(self):
        if not self.canceled:
            self.func(*self.args, **self.kwargs)
            self.completed = True

    def cancel(self):
        self.canceled = True
        if self.__handle != None:
            self.__handle.cancel()
